/*
 *  Copyright 2007 by Texas Instruments Incorporated.
 *  All rights reserved. Property of Texas Instruments Incorporated.
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 *
 *  @(#) TCP/IP_Network_Developers_Kit 1.92.00.22 01-10-2007 (ndk-b22)
 */
//--------------------------------------------------------------------------
// IP Stack
//--------------------------------------------------------------------------
// ti750.c - TI TLC16C750 Serial Port Driver with HDLC
//
// Author: Michael Denio
// Copyright 2001, 2003 by Texas Instruments, Inc.
//-------------------------------------------------------------------------
#include <std.h>
#include <sys.h>
#include <sem.h>
#include <hwi.h>
#include <c62.h>
#include <stkmain.h>
#include "llserial.h"
#include "ti752.h"

//
// We use the following direct access to registers
//
extern cregister volatile uint ICR;

//
// Declare our local functions
//
static void spInitialize( SDINFO *pi );
static void spStart( SDINFO *pi );
static void spStop( SDINFO *pi );
static void spIsr(SDINFO *pi);
static uint spWriteFifo( SDINFO *pi );

// Basic HDLC equates
#define HDLC_FLAGCHAR   0x7E    // Character to indicate frame boundary
#define HDLC_MINCHAR    0x20    // Minimum non-escaped char
#define HDLC_ESCCHAR    0x7D    // Character to indicate escape
#define HDLC_XORCHAR    0x20    // Escape XOR modifier

// FCS lookup table as calculated by the table generator.
#define PPPINITFCS16    0xffff  // Initial FCS value
#define PPPGOODFCS16    0xf0b8  // Good final FCS value
static const UINT16 fcstab[16] = {
   0x0000, 0x1081, 0x2102, 0x3183, 0x4204, 0x5285, 0x6306, 0x7387,
   0x8408, 0x9489, 0xA50A, 0xB58B, 0xC60C, 0xD68D, 0xE70E, 0xF78F
};

uint    SerRxGood=0,SerRxBadCRC=0;

//--------------------------------------------------------------------
// HwSerInit()
//
// Initialize serial environment and return instance count
//--------------------------------------------------------------------
uint HwSerInit()
{
    // EMIF control
    if( HW_CE_REG )
        *HW_CE_REG = HW_CE_VALUE;

#ifdef EVMDM642
    EVMDM642_OSD_init();
#endif

    return(2);
}

//--------------------------------------------------------------------
// HwSerShutdown()
//
// Shutdown Device Environment
//--------------------------------------------------------------------
void HwSerShutdown()
{
}

//--------------------------------------------------------------------
// HwSerOpen()
//
// Open Device Instance
//--------------------------------------------------------------------
uint HwSerOpen( SDINFO *pi )
{
    uint device;

    // Setup the base address
#ifdef EVMDM642
    if( !pi->PhysIdx )
    {
        pi->HwBaseAddr = ACE_PORT_BASE1;
        device = UARTA_IRQ;
    }
    else
    {
        pi->HwBaseAddr = ACE_PORT_BASE2;
        device = UARTB_IRQ;
    }
#endif

    // Attempt to reset the entire part
    spInitialize( pi );

    // Hook the interrupt
#ifdef EVMDM642
    EVMDM642_OSD_intHook( device, (EVMDM642_OSD_IntHandler)spIsr, pi );
#endif

    // Start
    spStart( pi );

    return(1);
}

//--------------------------------------------------------------------
// HwSerClose()
//
// Close Device Instance
//--------------------------------------------------------------------
void HwSerClose( SDINFO *pi )
{
    // Normally we'd shut off the part, but here we'll only
    // stop it. This is because it can't handle a soft reset.
    spStop( pi );
}

//--------------------------------------------------------------------
// HwSerSetConfig()
//
// Update hardware configuration
//--------------------------------------------------------------------
void HwSerSetConfig( SDINFO *pi )
{
    // Stop the part in preparation to re-initialize
    spStop( pi );

    // Attempt to reset the entire part
    spInitialize( pi );

    // Start
    spStart( pi );
}

//--------------------------------------------------------------------
// HwSerTxNext()
//
// Transmit Next Buffer on Tx Queue
//--------------------------------------------------------------------
void HwSerTxNext( SDINFO *pi )
{
    // Disable serial interrupts
    SREG(pi,UART_IER) = 0;

    if( !pi->TxFree )
    {
        SREG(pi,UART_IER) = TL16C752_IER_RX_AVAIL | TL16C752_IER_TX_READY;
        return;
    }

    pi->TxFree = 0;

    if( spWriteFifo( pi ) )
        SREG(pi,UART_IER) = TL16C752_IER_RX_AVAIL | TL16C752_IER_TX_READY;
    else
        SREG(pi,UART_IER) = TL16C752_IER_RX_AVAIL;
}

//--------------------------------------------------------------------
// HwSerPoll()
//
// Poll routine
//--------------------------------------------------------------------
void _HwSerPoll( SDINFO *pi,uint fTimerTick )
{
    (void)pi;
    (void)fTimerTick;
}

//--------------------------------------------------------------------
// spInitialize()
//
// Reset serial port and setup for operation
//--------------------------------------------------------------------
static void spInitialize( SDINFO *pi )
{
    uint        baudsetting;

    // Disable serial interrupts
    SREG(pi,UART_IER) = 0;

    // Calculate Baud setting based on BAUDRATENUM constant
    baudsetting = BAUDRATENUM / pi->Baud;

    // Set mode to baud latch enable and enable EFR
    SREG(pi,UART_LCR) = TL16C752_LCR_DLABEFR;

    // Set baud rate
    SREG(pi,UART_BRDL) = (UINT8)baudsetting;
    SREG(pi,UART_BRDH) = (UINT8)(baudsetting>>8);

    // Setup TX fifo trigger to 8 and RX fifo to 32
    SREG(pi,UART_EFR) = TL16C752_EFR_ENHANCE;
    SREG(pi,UART_TLR) = 0x82;  // 8*4=32 RX, 2*4=8 TX

    // Setup RX flow control in case its needed
    // Halt at 60, restore at 32
    SREG(pi,UART_TCR) = 0x8F;  // 8*4=32 Restore, 15*4=8 Halt

    // Enable FIFO, reset RX and TX
    SREG(pi,UART_FCR) = TL16C752_FCR_FIFO_ON |
                        TL16C752_FCR_RX_RESET | TL16C752_FCR_TX_RESET;

    // Set or disable flow control
    if( pi->FlowCtrl == HAL_SERIAL_FLOWCTRL_HARDWARE )
    {
        SREG(pi,UART_EFR) = TL16C752_EFR_ENHANCE |
                            TL16C752_EFR_RTSFLOW |
                            TL16C752_EFR_CTSFLOW;
        SREG(pi,UART_MCR) = TL16C752_MCR_IRQEN   |
                            TL16C752_MCR_DTR;
    }
    else
        SREG(pi,UART_MCR) = TL16C752_MCR_IRQEN   |
                            TL16C752_MCR_RTS     |
                            TL16C752_MCR_DTR;

    // Set mode (without latch enable)
    if( pi->Mode == HAL_SERIAL_MODE_7E1 )
        SREG(pi,UART_LCR) = 0x1a;      // 7E1
    else
        SREG(pi,UART_LCR) = 0x03;      // 8N1

    // Tansmitter is not free until we start
    pi->TxFree = 0;
}

//--------------------------------------------------------------------
// spStart()
//
// Start serial port Rx and Tx state machines
//--------------------------------------------------------------------
static void spStart( SDINFO *pi )
{
    // Enable RX and TX buffer interrupts
    SREG(pi,UART_IER) = TL16C752_IER_RX_AVAIL;
    pi->TxFree = 1;
}

//--------------------------------------------------------------------
// spStop()
//
// Stop serial port, abort current Rx and Tx
//--------------------------------------------------------------------
static void spStop( SDINFO *pi )
{
    // Disable serial interrupts
    SREG(pi,UART_IER) = 0;

    // Clear any pending interrupt
    ICR = HW_IFLAG;

    // Clear any waiting charmode data
    pi->CharReadIdx = pi->CharWriteIdx = pi->CharCount = 0;

    // Zap pending RX
    if( pi->hRxPend )
    {
        PBM_free( pi->hRxPend );
        pi->hRxPend = 0;
    }

    // Zap pending TX
    if( pi->hTxPend )
    {
        PBM_free( pi->hTxPend );
        pi->hTxPend = 0;
    }

    // Zap any queued transmits
    while( PBMQ_count(&pi->PBMQ_tx) )
        PBM_free( PBMQ_deq(&pi->PBMQ_tx) );

    // Tx not free
    pi->TxFree = 0;
}


//--------------------------------------------------------------------
// spWriteFifo()
//
// Writes out TX data to the TX FIFO. This function can be called
// in or out of the ISR.
//
// When the driver is open for HDLC operation, this code also frames,
// escapes, and validates the CRC for HDLC.
//
// Returns 1 if the system needs the TX FIFO Empty interrupt
//--------------------------------------------------------------------
static uint spWriteFifo( SDINFO *pi )
{
    uint   fifofree;
    UINT8  c;

    // We can write another FIFO set of bytes (less the ISR threshhold)
    fifofree = ACE_FIFO-8;

    while( fifofree )
    {
        // Start a new packet if we need to
        if( !pi->hTxPend )
        {
            pi->hTxPend = PBMQ_deq(&pi->PBMQ_tx);

            // If we don't have anything to send here, quit
            if( !pi->hTxPend )
            {
                if( fifofree == ACE_FIFO-8 )
                {
                    pi->TxFree = 1;
                    return(0);
                }
                return(1);
            }

            pi->pTxBuf  = PBM_getDataBuffer(pi->hTxPend) +
                          PBM_getDataOffset(pi->hTxPend);
            pi->TxCount = PBM_getValidLen(pi->hTxPend);

            // If in HDLC mode, setup for HDLC
            if( pi->hHDLC )
            {
                pi->TxCRC   = PPPINITFCS16;
                pi->TxFlag  = 1;
                pi->TxChar  = HDLC_FLAGCHAR;
            }
            else
                pi->TxFlag  = 0;
        }

        //
        // Continue with pending packet
        //

        // If inserting a character, do it now, else read from packet
        if( pi->TxFlag )
        {
            c = pi->TxChar;
            pi->TxFlag = 0;
        }
        else
        {
            c = *pi->pTxBuf++;
            pi->TxCount--;

            // If in HDLC mode, handle CRC and any "escape"
            if( pi->hHDLC )
            {
                // If we get here, we have a HDLC charater track CRC
                // We use 4 bit CRC to save table space
                pi->TxCRC ^= c;
                pi->TxCRC = (pi->TxCRC >> 4) ^ fcstab[0xF & pi->TxCRC];
                pi->TxCRC = (pi->TxCRC >> 4) ^ fcstab[0xF & pi->TxCRC];

                // See if we must "escape" this char
                if( (c < HDLC_MINCHAR && (pi->PeerMap & (1<<c)))
                    || c == HDLC_ESCCHAR || c == HDLC_FLAGCHAR )
                {
                    pi->TxFlag = 1;
                    pi->TxChar = c ^ HDLC_XORCHAR;
                    c = HDLC_ESCCHAR;
                }

                // If count is 2, then we can insert the checksum into
                // the packet
                if( pi->TxCount == 2 )
                {
                    pi->TxCRC = ~pi->TxCRC;
                    *pi->pTxBuf     = (UINT8)(pi->TxCRC%256);
                    *(pi->pTxBuf+1) = (UINT8)(pi->TxCRC/256);
                }
            }
        }

        // Write the character
        SREG(pi,UART_THR) = c;
        fifofree--;

        // See if we're done
        if( !pi->TxFlag && !pi->TxCount )
        {
            // In HDLC mode we may need to add one more flag byte
            if( pi->hHDLC && c != HDLC_FLAGCHAR )
            {
                pi->TxFlag  = 1;
                pi->TxChar  = HDLC_FLAGCHAR;
            }
            else
            {
                // Here were done with the packet
                PBM_free(pi->hTxPend);
                pi->hTxPend = 0;
            }
        }
    }

    return(1);
}

//--------------------------------------------------------------------
// spIsr()
//
// General purpose ISR function. Here we will read data from
// the device, or send more data to it. We also will signal the
// scheduler if ANY charmode data is available, or if a full
// HDLC buffer is ready.
//--------------------------------------------------------------------
static void spIsr( SDINFO *pi )
{
    uint   itype;
    uint   fSignal = 0;
    uint   fEnableTxInt;
    register UINT8  c;

    // Get this interrupt type
    itype = SREG(pi,UART_IIR) & 0xF;

    // Disable serial interrupts (this allow edge triggering to
    // detect a "stuck" interrupt.
    SREG(pi,UART_IER) = 0;

    // Check for receive data ready
    if( itype == TL16C752_IIR_RDA ||
        itype == TL16C752_IIR_TIMEOUT )
    {
        do
        {
            // Read the character
            c = SREG(pi,UART_RBR);

            // If in charmode, write this character to the charmode
            // buffer. Otherwise, we write it to the HDLC buffer.
            if( !pi->hHDLC )
            {
                if( pi->CharCount < CHAR_MAX )
                {
                    pi->CharBuf[pi->CharWriteIdx++] = c;
                    if( pi->CharWriteIdx == CHAR_MAX )
                        pi->CharWriteIdx = 0;
                    pi->CharCount++;
                    fSignal = 1;
                }
            }
            else
            {
                // If we don't have a packet ready, get one
                if( !pi->hRxPend )
                {
                    pi->hRxPend = PBM_alloc(RX_MAX+PKT_PREPAD);

                    // If we don't get a buffer, toss the character
                    if( !pi->hRxPend )
                        continue;
InitRxPacket:
                    PBM_setDataOffset(pi->hRxPend,PKT_PREPAD);
                    pi->pRxBuf  = PBM_getDataBuffer(pi->hRxPend)+PKT_PREPAD;
                    pi->RxCount = 0;
                    pi->RxCRC   = PPPINITFCS16;
                    pi->RxFlag  = 0;
                }


                // If its a marker char, act on it
                if( c == HDLC_FLAGCHAR )
                {
                    // If we have some data in our HDLC packet, try and validate
                    // it. Otherwise, ignore this packet
                    if( !pi->RxCount )
                        continue;

                    // Verify the packet CRC
                    if( pi->RxCRC != PPPGOODFCS16 )
                    {
                        SerRxBadCRC++;
                        goto InitRxPacket;
                    }
                    else
                    {
                        // Correct the packet size and offset to give to PPP
                        SerRxGood++;
                        PBM_setValidLen(pi->hRxPend,pi->RxCount);
                        PBM_setIFRx(pi->hRxPend,pi->hHDLC);
                        PBMQ_enq( &pi->PBMQ_rx, pi->hRxPend );
                        pi->hRxPend = 0;
                        fSignal = 1;
                    }
                    continue;
                }

                // If this is an escape char, enter escape mode
                if( c == HDLC_ESCCHAR )
                {
                    pi->RxFlag = 1;
                    continue;
                }

                // Don't receive too much
                if( pi->RxCount == RX_MAX )
                    continue;

                // If we are in the middle of an "escape", then we convert
                // this character
                if( pi->RxFlag )
                {
                    pi->RxFlag = 0;
                    c ^= HDLC_XORCHAR;
                }

                // Track the CRC so we can verify it later
                pi->RxCRC ^= c;
                pi->RxCRC = (pi->RxCRC >> 4) ^ fcstab[0xF & pi->RxCRC];
                pi->RxCRC = (pi->RxCRC >> 4) ^ fcstab[0xF & pi->RxCRC];

                *pi->pRxBuf++ = c;
                pi->RxCount++;
            }
        } while( SREG(pi,UART_LSR) & TL16C752_LSR_DR );
    }

    // Check/Enable transmit FIFO empty int
    if( pi->TxFree )
        fEnableTxInt = 0;
    else if( itype == TL16C752_IIR_THRE )
        fEnableTxInt = spWriteFifo( pi );
    else
        fEnableTxInt = 1;

    // Enable serial interrupts
    if( fEnableTxInt )
        SREG(pi,UART_IER) = TL16C752_IER_RX_AVAIL | TL16C752_IER_TX_READY;
    else
        SREG(pi,UART_IER) = TL16C752_IER_RX_AVAIL;

    if( fSignal )
        STKEVENT_signal( pi->hEvent, STKEVENT_SERIAL, 1 );
}


